# cython: wraparound=False, nonecheck=False, boundscheck=False, cdivision=True, initializedcheck=False

"""
The cython module which contains the internal simulator algorithms.

This is not intended to be interacted with directly.
It is intended for the user to only interact with the class :any:`Simulation`.
"""

''' Use these commands to enable line tracing.
# cython: linetrace=True
# cython: binding=True
# distutils: define_macros=CYTHON_TRACE_NOGIL=1
'''

from libc.math cimport log, lgamma, sqrt
from libc.stdint cimport int64_t, uint8_t, uint32_t
cimport cython
from numpy cimport npy_intp
import numpy as np
cimport numpy as np
from cpython.pycapsule cimport PyCapsule_GetPointer
from numpy.random cimport bitgen_t
from numpy.random cimport BitGenerator
from numpy.random import PCG64
from numpy.random.c_distributions cimport \
    (random_hypergeometric, random_interval, random_multinomial, random_geometric, binomial_t)
import time


cdef class Simulator:
    """Base class for the algorithm that runs the simulation.

    The configuration is stored as an array of size q, so the states are the indices 0, ..., q-1.

    Attributes:
        config: The integer array of counts representing the current configuration.
        n: The population size (sum of config).
        t: The current number of elapsed interaction steps.
        q: The total number of states (length of config).
        is_random: A boolean that is true if there are any random transitions.
        random_depth: The largest number of random outputs from any random transition.
        gen: A BitGenerator object which is the pseudorandom number generator.
        bitgen: A pointer to the BitGenerator, needed to for the numpy random C-API.
    """
    cdef public int64_t [::1] config
    cdef public int64_t n, t
    cdef npy_intp q
    cdef public npy_intp [:,:,:,] delta
    cdef public uint8_t [:,:] null_transitions
    cdef uint8_t is_random
    cdef npy_intp [:,:,:] random_transitions
    cdef npy_intp [:,:] random_outputs
    cdef double [::1] transition_probabilities
    cdef npy_intp random_depth
    cdef BitGenerator gen
    cdef bitgen_t *bitgen

    # random_outputs being typed causes an error when the protocol is not random and the array is empty.
    def __init__(self, int64_t [::1] init_array, npy_intp [:,:,:] delta, uint8_t [:,:] null_transitions,
                 npy_intp [:,:,:] random_transitions, random_outputs,
                 double [::1] transition_probabilities, seed=None):
        """Initializes the main data structures for :any:`Simulator`.

        Args:
            init_array: An integer array of counts representing the initial configuration.
            delta: A q x q x 2 array representing the transition function.
                Delta[i, j] gives contains the two output states.
            null_transitions: A q x q boolean array where entry [i, j] says if these states have a null interaction.
            random_transitions: A q x q x 2 array. Entry [i, j, 0] is the number of possible outputs if
                transition [i, j] is random, otherwise it is 0. Entry [i, j, 1] gives the starting index to find
                the outputs in the array random_outputs if it is random.
            random_outputs: An array containing all outputs of random transitions, whose indexing information
                is contained in random_transitions.
            transition_probabilities: An array containing all random transition probabilities, whose indexing matches
                random_outputs.
            seed (optional): An integer seed for the pseudorandom number generator.
        """
        self.config = init_array
        self.n = sum(init_array)
        self.q = len(init_array)
        self.delta = delta
        self.null_transitions = null_transitions
        self.is_random = np.any(random_transitions)
        if self.is_random:
            self.random_depth = 0
            for i in range(self.q):
                for j in range(self.q):
                    self.random_depth = max(self.random_depth, random_transitions[i, j, 0])
            self.random_transitions = random_transitions
            self.random_outputs = random_outputs
            self.transition_probabilities = transition_probabilities
        self.t = 0
        self.gen = PCG64(seed=seed)
        self.bitgen = <bitgen_t *> PyCapsule_GetPointer(self.gen.capsule, "BitGenerator")

    def run(self, int64_t num_steps, double max_wallclock_time = 60 * 60):
        """Base function which will be called to run the simulation for a fixed number of steps.

        Args:
            num_steps: The number of steps to run the simulation.
            max_wallclock_time: A bound in seconds on how long the simulator will run for.
        """
        pass

    def reset(self, int64_t [::1] config, int64_t t = 0):
        """Base function which will be called to reset the simulation.

        Args:
            config: The configuration array to reset to.
            t: The new value of self.t. Defaults to 0.
        """
        pass


cdef class SimulatorSequentialArray(Simulator):
    """A Simulator that sequentially chooses random agents from an array.

    Attributes:
        population: A length-n array with entries in 0, ..., q-1 giving the states of each agent.
    """
    cdef npy_intp [::1] population

    def __init__(self, *args):
        """Initializes Simulator, then creates the population array."""
        Simulator.__init__(self, *args)
        self.make_population()

    def make_population(self):
        """Creates the array self.population.

        This is an array of agent states, where the count of each state comes from :any:`Simulator.config`.
        """
        self.population = np.zeros(self.n, dtype=np.intp)
        cdef npy_intp i, j, k = 0
        for i in range(self.q):
            for j in range(self.config[i]):
                self.population[k] = i
                k += 1

    def run(self, int64_t end_step, double max_wallclock_time = 60 * 60):
        """Samples random pairs of agents and updates them until reaching end_step."""
        cdef npy_intp i, j, k, a, b
        cdef double u
        cdef double [:] ps
        cdef double end_time = time.perf_counter() + max_wallclock_time
        while self.t < end_step and time.perf_counter() < end_time:
            # random integer in [0, ... , n-1]
            i = random_interval(self.bitgen, self.n - 1)
            j = random_interval(self.bitgen, self.n - 1)
            while i == j:
                # rejection sampling to quickly get distinct pair
                j = random_interval(self.bitgen, self.n - 1)
            a, b = self.population[i], self.population[j]
            if not self.null_transitions[a, b]:
                if self.is_random and self.random_transitions[a, b, 0]:
                    k = self.random_transitions[a, b, 1]
                    # sample from a probability distribution contained in [k, k+1, ...]
                    u = self.bitgen.next_double(self.bitgen.state) - self.transition_probabilities[k]
                    while u > 0:
                        k += 1
                        u -= self.transition_probabilities[k]
                    self.population[i], self.population[j] = self.random_outputs[k]
                else:
                    self.population[i], self.population[j] = self.delta[a, b]
                self.config[a] -= 1
                self.config[b] -= 1
                self.config[self.population[i]] += 1
                self.config[self.population[j]] += 1
            self.t += 1
        return self.config

    def reset(self, int64_t [::1] config, int64_t t = 0):
        """Reset to a given configuration.

        Sets all parameters necessary to change the configuration.

        Args:
            config: The configuration array to reset to.
            t: The new value of :any:`t`. Defaults to 0.
        """
        self.config = config
        self.t = t
        self.n = sum(config)
        self.make_population()


cdef class SimulatorMultiBatch(Simulator):
    """Uses the MultiBatch algorithm to simulate O(sqrt(n)) interactions in parallel.

    The MultiBatch algorithm comes from the paper
    *Simulating Population Protocols in Subconstant Time per Interaction*
    (https://arxiv.org/abs/2005.03584).
    Beyond the methods described in the paper, this class also dynamically switches
    to Gillespie's algorithm when the number of null interactions is high.

    Attributes:
        urn: An :any:`Urn` object which stores the configuration and has methods for sampling.
        updated_counts: An additional :any:`Urn` where agents are stored that have been
            updated during a batch.
        logn: Precomputed log(n).
        batch_threshold: Minimum number of interactions that must be simulated in each
            batch. Collisions will be repeatedly sampled up until batch_threshold
            interaction steps, then all non-colliding pairs of 'delayed agents' are
            processed in parallel.
        row_sums: Array which stores sampled counts of initiator agents
            (row sums of the 'D' matrix from the paper).
        row: Array which stores the counts of responder agents for each type of
            initiator agent (one row of the 'D' matrix from the paper).
        m: Array which holds the outputs of samples from a multinomial distribution
            for batch random transitions.
        do_gillespie: A boolean determining if the we are currently doing Gillespie steps.
        silent: A boolean determining if the configuration is silent
            (all interactions are null).
        reactions: A (num_reactions) x 4 array giving a list of reactions,
            as [input input output output]
        enabled_reactions: An array holding indices of all currently applicable reactions.
        num_enabled_reactions: The number of meaningful indices in enabled_reactions.
        propensities: A num_reactions x 1 array holding the propensities of each reaction.
            The propensity of a reaction is the probability of that reaction * (n choose 2).
        reaction_probabilities: A num_reactions x 1 array giving the probability of each
            reaction, given that those two agents interact.
        gillespie_threshold: The probability of a non-null interaction must be below this
            threshold to keep doing Gillespie steps.
        coll_table: Precomputed values to speed up the function sample_coll(r, u).
        coll_table_r_values: Values of r, giving one axis of coll_table.
        coll_table_u_values: Values of u, giving the other axis of coll_table.
        num_r_values: len(coll_table_r_values), first axis of coll_table.
        num_u_values: len(coll_table_u_values), second axis of coll_table.
        r_constant: Used in definition of coll_table_r_values.
    """

    cdef public Urn urn
    cdef Urn updated_counts
    cdef double logn
    cdef int64_t batch_threshold
    cdef int64_t [::1] row_sums, row, m
    cdef public bint do_gillespie, silent
    cdef public npy_intp [:,:] reactions
    cdef public npy_intp [:] enabled_reactions
    cdef public npy_intp num_enabled_reactions
    cdef double [::1] propensities
    cdef public double [::1] reaction_probabilities
    cdef double gillespie_threshold
    cdef binomial_t _binomial  # needed to interface with the numpy multinomial C function
    cdef int64_t [:, :] coll_table
    cdef int64_t [::1] coll_table_r_values
    cdef double [::1] coll_table_u_values
    cdef npy_intp num_u_values
    cdef npy_intp num_r_values
    cdef int64_t r_constant

    def __init__(self, *args):
        """Initializes all additional data structures needed for MultiBatch Simulator."""

        Simulator.__init__(self, *args)

        self.set_n_parameters()

        self.urn = Urn.create(self.config, self.bitgen)
        self.updated_counts = Urn.create(np.zeros(self.q, dtype=np.int64), self.bitgen)
        self.row_sums = np.zeros(self.q, dtype=np.int64)
        self.row = np.zeros(self.q, dtype=np.int64)
        self.m = np.zeros(self.random_depth, dtype=np.int64)
        self.silent = False
        self.do_gillespie = False

        # enumerate reactions for gillespie
        reactions = []
        reaction_probabilities = []
        cdef npy_intp i, j, k, a, b
        cdef double p
        for i in range(self.q):
            for j in range(i+1):
                # check if interaction is symmetric
                symmetric = False
                if sorted(np.asarray(self.delta[i,j])) == sorted(np.asarray(self.delta[j, i])):
                    if self.is_random and self.random_transitions[i, j, 0] == self.random_transitions[j, i, 0] > 0:
                        a, b = self.random_transitions[i, j, 1], self.random_transitions[j, i, 1]
                        symmetric = True
                        for k in range(self.random_transitions[i, j, 0]):
                            if sorted(np.asarray(self.random_outputs[a + k])) != \
                                    sorted(np.asarray(self.random_outputs[b + k])):
                                symmetric = False
                    else:
                        symmetric = True
                if symmetric:
                    indices = [(i, j, 1.)]
                # if interaction is not symmetric, each distinct order gets added as reactions with half proability
                else:
                    indices = [(i, j, 0.5), (j, i, 0.5)]
                for a, b, p in indices:
                    if not self.null_transitions[a, b]:
                        if self.is_random and self.random_transitions[a, b, 0]:
                            for k in range(self.random_transitions[a, b, 0]):
                                output = list(self.random_outputs[self.random_transitions[a, b, 1] + k])
                                if output != [a, b]:
                                    reactions.append([a, b] + output)
                                    reaction_probabilities.append(
                                        self.transition_probabilities[self.random_transitions[a, b, 1] + k] * p)
                        else:
                            reactions.append([a, b] + list(self.delta[a, b]))
                            reaction_probabilities.append(p)
        self.reactions = np.array(reactions, dtype = np.intp)
        self.reaction_probabilities = np.array(reaction_probabilities, dtype = float)
        self.propensities = np.zeros(len(self.reactions), dtype = float)
        self.enabled_reactions = np.zeros(len(self.reactions), dtype = np.intp)
        self.get_enabled_reactions()

    def set_n_parameters(self):
        """Initialize all parameters that depend on the population size n."""
        self.logn = log(self.n)
        # theoretical optimum for batch_threshold is Theta(sqrt(n / logn) * q) agents / batch
        self.batch_threshold = int(min(sqrt(self.n / self.logn) * self.q, self.n ** 0.7))
        # first rough approximation for probability of successful reaction where we want to do gillespie
        self.gillespie_threshold = 2 / sqrt(self.n)

        # build table for precomputed coll(n, r, u) values
        self.num_r_values = int(10 * log(self.n))
        self.num_u_values = int(10 * log(self.n))
        self.r_constant = max(int(1.5 * self.batch_threshold) // ((self.num_r_values - 2) ** 2), 1)
        self.coll_table = np.zeros((self.num_r_values, self.num_u_values), dtype=np.int64)
        self.coll_table_r_values = np.array([2 + self.r_constant * (i ** 2) for i in range(self.num_r_values - 1)]
                                            + [self.n], dtype=np.int64)
        self.coll_table_u_values = np.linspace(0, 1, self.num_u_values, dtype=float)
        cdef npy_intp i, j
        for i in range(self.num_r_values):
            for j in range(self.num_u_values):
                self.coll_table[i, j] = self.sample_coll(self.coll_table_r_values[i],
                                                         self.coll_table_u_values[j], has_bounds=False)

    def reset(self, int64_t [::1] config, int64_t t = 0):
        """Reset to a given configuration.

        Sets all parameters necessary to change the configuration.

        Args:
            config: The configuration array to reset to.
            t: The new value of :any:`t`. Defaults to 0.
        """
        self.config = config
        self.urn = Urn.create(self.config, self.bitgen)
        cdef int64_t n = sum(config)
        if n != self.n:
            self.n = n
            self.set_n_parameters()
        self.t = t
        self.silent = False
        self.do_gillespie = False

    def run(self, int64_t end_step, double max_wallclock_time = 60 * 60):
        """Run the simulation for a fixed number of steps.

        Args:
            end_step: Will run until self.t = end_step.
            max_wallclock_time: A bound in seconds this will run for.
        """
        cdef double end_time = time.perf_counter() + max_wallclock_time
        while self.t < end_step and time.perf_counter() < end_time:
            if self.silent:
                self.t = end_step
                return
            elif self.do_gillespie:
                self.gillespie_step(end_step)
            else:
                self.multibatch_step(end_step)

    def run_until_silent(self, int64_t [::1] config):
        """Run the simulation until silent."""
        while not self.silent:
            if self.do_gillespie:
                self.gillespie_step()
            else:
                self.multibatch_step()

    cdef (npy_intp, npy_intp) unordered_delta(self, npy_intp a, npy_intp b):
        """Chooses sender/receiver, then applies delta to input states a, b."""
        cdef double u
        cdef npy_intp k
        cdef uint32_t coin = self.bitgen.next_uint32(self.bitgen.state) & 1
        # swap roles of a, b and swap return order by using indices coin, 1-coin
        if coin:
            b, a = a, b
        if self.is_random and self.random_transitions[a, b, 0]:
            # find the appropriate random output by linear search
            k = self.random_transitions[a, b, 1]
            u = self.bitgen.next_double(self.bitgen.state) - self.transition_probabilities[k]
            while u > 0:
                k += 1
                u -= self.transition_probabilities[k]
            return self.random_outputs[k][coin], self.random_outputs[k][1-coin]
        else:
            return self.delta[a, b, coin], self.delta[a, b, 1-coin]

    def get_enabled_reactions(self):
        """Updates :any:`enabled_reactions` and :any:`num_enabled_reactions`."""
        cdef npy_intp i, reactant_1, k
        self.num_enabled_reactions = 0
        for i in range(len(self.reactions)):
            reactant_1, reactant_2 = self.reactions[i][0], self.reactions[i][1]
            if (reactant_1 == reactant_2 and self.config[reactant_1] >= 2) or \
                    (reactant_1 != reactant_2 and self.config[reactant_1] >= 1 and self.config[reactant_2] >= 1):
                self.enabled_reactions[self.num_enabled_reactions] = i
                self.num_enabled_reactions += 1

    def gillespie_step(self, int64_t t_max = 0):
        """Samples the time until the next non-null interaction and updates.

        Args:
            t_max: Defaults to 0.
                If positive, the maximum value of :any:`t` that will be reached.
                If the sampled time is greater than t_max, then it will instead
                be set to t_max and no reaction will be performed.
                (Because of the memoryless property of the geometric, this gives a
                faithful simulation up to step t_max).
        """

        cdef npy_intp [:] r
        cdef double total_propensity = self.get_total_propensity()
        if total_propensity == 0:
            self.silent = True
            return
        cdef double n = self.n
        cdef double success_probability = total_propensity / (n * (n-1) / 2)
        cdef bint enabled_reactions_changed = False

        if success_probability > self.gillespie_threshold:
            self.do_gillespie = False
        # add a geometric number of steps, based on success probability
        new_t = self.t + random_geometric(self.bitgen, success_probability)
        self.t += random_geometric(self.bitgen, success_probability)
        # if t_max was exceeded, stop at step t_max without performing a reaction
        if self.t > t_max > 0:
            self.t = t_max
            return
        # sample the successful reaction r, currently just using linear search
        x = self.bitgen.next_double(self.bitgen.state) * total_propensity
        cdef npy_intp i = 0
        while x > 0:
            x -= self.propensities[self.enabled_reactions[i]]
            i += 1
        r = self.reactions[self.enabled_reactions[i - 1]]
        # updated with the successful reaction r
        # if any products were not already present, will updated enabled_reactions
        if self.config[r[2]] == 0 or self.config[r[3]] == 0:
            enabled_reactions_changed = True
        ## this is a bit wasteful, but want to make sure the urn data structure stays intact
        self.urn.add_to_entry(r[0], -1)
        self.urn.add_to_entry(r[1], -1)
        self.urn.add_to_entry(r[2], 1)
        self.urn.add_to_entry(r[3], 1)
        # if any reactants are now absent, will updated enabled_reactions
        if enabled_reactions_changed or self.config[r[0]] == 0 or self.config[r[1]] == 0:
            self.get_enabled_reactions()

    cpdef double get_total_propensity(self):
        """Calculates the probability the next interaction is non-null."""
        cdef npy_intp i, j
        # make sure these are all doubles, because they will be squared and could overflow int64_t
        cdef double a, b
        cdef double total_propensity = 0
        for j in range(self.num_enabled_reactions):
            i = self.enabled_reactions[j]
            a, b = self.config[self.reactions[i][0]], self.config[self.reactions[i][1]]
            if self.reactions[i][0] == self.reactions[i][1]:
                self.propensities[i] = (a * (a-1) / 2) * self.reaction_probabilities[i]
            else:
                self.propensities[i] = a * b * self.reaction_probabilities[i]
            total_propensity += self.propensities[i]
        return total_propensity

    cpdef void multibatch_step(self, int64_t t_max = 0):
        """Sample collisions to build a batch, then update the entire batch in parallel.

        See the paper for a more detailed explanation of the algorithm.
        """
        cdef int64_t num_delayed, l
        cdef double t1, t2, t3, u, r, end_step
        cdef npy_intp a, b, c, i, j, i_max, j_max, o_i, o_j


        self.updated_counts.reset()
        self.updated_counts.order = self.urn.order
        # start with count 2 of delayed agents (guaranteed for the next interaction)
        num_delayed = 2

        t1 = time.perf_counter()
        # batch will go for at least batch_threshold interactions, unless passing t_max
        end_step = self.t + self.batch_threshold
        if t_max > 0:
            end_step = min(end_step, t_max)
        while self.t + num_delayed // 2 < end_step:
            u = self.bitgen.next_double(self.bitgen.state)
            l = self.sample_coll(r = num_delayed + self.updated_counts.size,
                                 u = u, has_bounds=True)
            # add (l-1) // 2 pairs of delayed agents, the lth agent a was already picked, so has a collision
            num_delayed += 2 * ((l-1) // 2)

            # If the sampled collision happens after t_max, then include delayed agents up until t_max
            #   and do not perform the collision.
            if self.t + num_delayed // 2 >= t_max > 0:
                num_delayed = (t_max - self.t) * 2
                break

            # sample if a was a delayed or an updated agent
            u = self.bitgen.next_double(self.bitgen.state)
            r = num_delayed / (num_delayed + self.updated_counts.size)
            # true with probability num_delayed / (num_delayed + num_updated)
            if u * (num_delayed + self.updated_counts.size) <= num_delayed:
                # if a was delayed, need to first update a with its first interaction before the collision
                # c is the delayed partner that a interacted with, so add this interaction
                a = self.urn.sample_one()
                c = self.urn.sample_one()
                a, c = self.unordered_delta(a,c)
                self.t += 1
                # c is moved from delayed to updated, a is currently uncounted
                self.updated_counts.add_to_entry(c, 1)
                num_delayed -= 2
            else:
                # if a was updated, we simply sample a and remove it from updated counts
                a = self.updated_counts.sample_one()

            if l % 2 == 0:  # when l is even, the collision must with with a formally untouched agent
                b = self.urn.sample_one()
            else: # when l is odd, the collision is with the next agent, either untouched, delayed, or updated
                u = self.bitgen.next_double(self.bitgen.state)
                if u * (self.n - 1) < self.updated_counts.size:
                    # b is an updated agent, simply remove it
                    b = self.updated_counts.sample_one()
                else:
                    # we simply remove b from C is b is untouched
                    b = self.urn.sample_one()
                    # if b was delayed, we have to do the past interaction
                    if u * (self.n - 1) < self.updated_counts.size + num_delayed:
                        c = self.urn.sample_one()
                        b, c = self.unordered_delta(b,c)
                        self.t += 1
                        self.updated_counts.add_to_entry(c, 1)
                        num_delayed -= 2

            a, b = self.unordered_delta(a,b)
            self.t += 1
            self.updated_counts.add_to_entry(a, 1)
            self.updated_counts.add_to_entry(b, 1)

        t2 = time.perf_counter()
        self.do_gillespie = True  # if entire batch are null reactions, stays true and switches to gillspie
        i_max = self.urn.sample_vector(num_delayed // 2, self.row_sums)
        for i in range(i_max):
            o_i = self.urn.order[i]
            j_max = self.urn.sample_vector(self.row_sums[o_i], self.row)
            for j in range(j_max):
                o_j = self.urn.order[j]
                if self.is_random and self.random_transitions[o_i, o_j, 0]:
                    # don't switch to gillespie because we did a random transition
                    # TODO: this might not switch to gillespie soon enough in certain cases
                    self.do_gillespie = False
                    a = self.random_transitions[o_i, o_j, 0]  # better written using walrus operator
                    b = self.random_transitions[o_i, o_j, 1]
                    # updates the first a entries of m to hold a multinomial,
                    # giving the number of times for each random transition
                    self.m[:] = 0
                    random_multinomial(self.bitgen, self.row[o_j], &self.m[0], &self.transition_probabilities[b],
                                       a, &self._binomial)
                    for c in range(a):
                        self.updated_counts.add_to_entry(self.random_outputs[b+c,0], self.m[c])
                        self.updated_counts.add_to_entry(self.random_outputs[b+c,1], self.m[c])
                else:
                    if self.do_gillespie:
                        # if transition is non-null, we will set do_gillespie = False
                        self.do_gillespie = self.null_transitions[o_i, o_j]
                    # We are directly adding to updated_counts.config rather than using the function
                    #   updated_counts.add_to_entry for speed. None of the other urn features of updated_counts will
                    #   be used until it is reset in the next loop, so this is fine.
                    self.updated_counts.config[self.delta[o_i, o_j, 0]] += self.row[o_j]
                    self.updated_counts.config[self.delta[o_i, o_j, 1]] += self.row[o_j]

        self.t += num_delayed // 2
        # TODO: this is the only part scaling when the number of states (but not reached states) blows up
        self.urn.add_vector(self.updated_counts.config)

        t3 = time.perf_counter()

        # Dynamically update batch threshold, by comparing the times t2 - t1 of the collision sampling and
        #   the time t_3 - t_2 of the batch processing. Batch_threshold is adjusted to try to ensure
        #   t_2 - t_1 = t_3 - t_2
        self.batch_threshold = int(((t3 - t2) / (t2 - t1)) ** 0.1 * self.batch_threshold)
        # Keep the batch threshold within some fixed bounds.
        self.batch_threshold = min(self.batch_threshold, 2 * self.n // 3)
        self.batch_threshold = max(self.batch_threshold, 3)

        self.urn.sort()

        # update enabled_reactions if switching to gillespie
        if self.do_gillespie:
            self.get_enabled_reactions()

    cdef int64_t sample_coll(self, int64_t r, double u, bint has_bounds=True):
        """Returns a sample l ~ coll(n, r) from the collision length distribution.
        
        See Lemma 3 in the source paper. The distribution gives the number of agents needed to pick an agent twice,
        when r unique agents have already been selected.
        Inversion sampling with binary search is used, based on the formula 
            P(l > t) = (n - r)! / (n - r - t)! / (n^t).
        We sample a uniform random variable u, and find the value t such that 
            P(l > t) < U < P(l > t - 1). 
        Taking logarithms and using the lgamma function, this required formula becomes
            P(l > t) < U
             <-->
            lgamma(n - r + 1) - lgamma(n - r - t + 1) - t * log(n) < log(u).
        We will do binary search with bounds t_lo, t_hi that maintain the invariant
            P(l > t_hi) < U and P(l > t_lo) >= U.
        Once we get t_lo = t_hi - 1, we can then return t = t_hi as the output.
        
        A value of fixed outputs for u, r will be precomputed, which gives a lookup table for starting values
        of t_lo, t_hi. This function will first get called to give coll(n, r_i, u_i) for a fixed range of values 
        r_i, u_i. Then actual samples of coll(n, r, u) will find values r_i <= r < r_{i+1} and u_j <= u < u_{j+1}.
        By monotonicity in u, r, we can then set t_lo = coll(n, r_{i+i}, u_{j+1}) and t_hi = coll(n, r_i, u_j).
        
        Args:
            r: The number of agents which have already been chosen.
            u: A uniform random variable.
            has_bounds: Has the table for precomputed values of r, u already been computed?
                (This will be false while the function is being called to populate the table.)
            
        Returns:
            The number of sampled agents to get the first collision (including the collided agent).
        """
        cdef int64_t t_lo, t_mid, t_hi
        cdef npy_intp i, j
        cdef double logu, lhs
        logu = log(u)
        lhs = lgamma(self.n-r+1) - logu
        # The condition P(l < t) < U becomes
        #     lhs < lgamma(n - r - t + 1) + t * log(n)

        if has_bounds:
            # Look up bounds from coll_table.

            # For r values, we invert the definition of self.coll_table_r_values:
            #   np.array([2 + self.r_constant * (i ** 2) for i in range(self.num_r_values - 1)] + [self.n])
            i = int(sqrt((r - 2) / self.r_constant))
            i = min(i, self.num_r_values - 2)

            # for u values we similarly invert the definition: np.linspace(0, 1, num_u_values)
            j = int(u * (self.num_u_values - 1))

            # assert self.coll_table_r_values[i] <= r <= self.coll_table_r_values[i+1]
            # assert self.coll_table_u_values[j] <= u <= self.coll_table_u_values[j+1]
            t_lo = self.coll_table[i + 1, j + 1]
            t_hi = min(self.coll_table[i, j], self.n - r + 1)
        else:
            # When building the table, we start with bounds that always hold.
            if r >= self.n:
                return 1
            t_lo = 0
            t_hi = self.n - r

        # We maintain the invariant that P(l > t_lo) >= u and P(l > t_hi) < u
        # Equivalently, lhs >= lgamma(n - r - t_lo + 1) + t_lo * logn and
        #               lhs <  lgamma(n - r - t_hi + 1) + t_hi * logn
        while t_lo < t_hi - 1:
            t_mid = (t_lo + t_hi) // 2
            if lhs < lgamma(self.n - r + 1 - t_mid) + t_mid * self.logn:
                t_hi = t_mid
            else:
                t_lo = t_mid
        return t_hi


cdef class Urn:
    """Data structure for a multiset that supports fast random sampling.

    Attributes:
        config: The integer array giving the counts of the multiset.
        bitgen: Pointer to a BitGenerator, needed for numpy random C API.
        order: An integer array giving the ranking of the counts from largest to smallest.
        size: sum(config).
        length: len(config).
    """

    cdef public int64_t [::1] config
    cdef bitgen_t *bitgen
    cdef npy_intp [::1] order
    cdef int64_t size
    cdef npy_intp length

    @staticmethod
    cdef Urn create(int64_t [::1] config, bitgen_t * bitgen):
        """Initializes a new Urn object.
        
        Args:
            config: The configuration array for the urn.
            bitgen: The BitGenerator pointer. Because this is not a Python object, 
                we cannot using the typical __init__ constructor. 
                Calling this create method is a workaround.
        """
        cdef Urn urn = Urn.__new__(Urn)
        urn.config = config
        urn.bitgen = bitgen
        urn.size = sum(config)
        urn.length = len(config)
        urn.order = np.array(range(len(config)), dtype=np.intp)
        urn.sort()
        return urn

    cdef void sort(self):
        """Updates self.order.
        
        Uses insertion sort to maintain that 
            config[order[0]] >= config[order[1]] >= ... >= config[order[q]].
        This method is used to have O(q) time when order is almost correct.
        """
        cdef npy_intp i, j, k, o_i

        for i in range(1,len(self.config)):
            # See if the entry at self.order[i] needs to be moved earlier.
            # Recursively, we have ensured that order[0], ..., order[i-1] have the correct order.
            o_i = self.order[i]
            # j will be the index where self.order[i] should be inserted to.
            j = i
            while j > 0 and self.config[o_i] > self.config[self.order[j-1]]:
                j -= 1
            # Index at order[i] will get moved to order[j], and all indices order[j], ..., order[i-1] get right shifted
            # First do the right shift, moving order[i-k] for k = 1, ..., i-j
            for k in range(1, i-j+1):
                self.order[i + 1 - k] = self.order[i - k]
            self.order[j] = o_i

    cdef npy_intp sample_one(self):
        """Samples and removes one element, returning its index.
        
        Returns:
            The index of the random sample from the urn.
        """
        cdef npy_intp index, i=0
        cdef int64_t x = random_interval(self.bitgen, self.size - 1)
        while x >= 0:
            index = self.order[i]
            x -= self.config[index]
            i += 1
        self.config[index] -= 1
        self.size -= 1
        return index

    cdef void add_to_entry(self, npy_intp index, int64_t amount = 1):
        """Adds one element at index.
        
        Args:
            index: The index to add to.
            amount: The integer amount to add / subtract.
        """
        self.config[index] += amount
        self.size += amount

    cdef npy_intp sample_vector(self, int64_t n, int64_t [::1] v):
        """Samples n elements, returning them as a vector.
        
        Args:
            n: number of elements to sample
            v: the array to write the output vector in
                (this is faster than re-initializing an output array)
            
        Returns:
            nz: the number of nonzero entries
                v[self.order[i]] for i in range(nz) can then loop over only 
                    the nonzero entries of the vector
        """
        cdef int64_t init_n = n
        cdef npy_intp index, entries, i = 0
        cdef int64_t total = self.size
        cdef int64_t h
        v[:] = 0
        while n > 0 and i < self.length - 1:
            index = self.order[i]
            total -= self.config[index]
            h = random_hypergeometric(self.bitgen, self.config[index], total, n)
            v[index] = h
            n -= h
            self.size -= h
            self.config[index] -= h
            i += 1
        if n:
            v[self.order[i]] = n
            self.config[self.order[i]] -= n
            self.size -= n
            i += 1
        return i

    cdef void add_vector(self, int64_t [::1] vector):
        """Adds a vector of elements to the urn.
        
        Args:
            vector: An integer vector to add to the urn.
        """
        cdef npy_intp i = 0
        for i in range(self.length):
            self.config[i] += vector[i]
            self.size += vector[i]

    cdef void reset(self):
        """Set the counts back to zero."""
        self.config[:] = 0
        self.size = 0



"""
A module for :any:`Snapshot` objects used to visualize the protocol during or after
the simulation has run.

:any:`Snapshot` is a base class for snapshot objects that get are updated by :any:`Simulation`.

:any:`Plotter` is a subclass of :any:`Snapshot` that creates a matplotlib figure and axis.
It also gives the option for a state_map function which maps states to the categories which
will show up in the plot.

:any:`StatePlotter` is a subclass of :any:`Plotter` that creates a barplot of the counts
in categories.

:any:`HistoryPlotter` is a subclass of :any:`Plotter` that creates a lineplot of the counts
in categories over time.
"""

from typing import Optional, Callable, Hashable, Any

from natsort import natsorted
import numpy as np
import pandas as pd  # type: ignore

State = Hashable


class Snapshot:
    """Base class for snapshot objects.

    Attributes:
        simulation: The :any:`Simulation` object that initialized and will update the
            :any:`Snapshot`.
            This attribute gets set when the :any:`Simulation` object calls
            :any:`add_snapshot`.
        update_time: How many seconds will elapse between calls to update while
            in the :any:`Simulation.run` method of :any:`simulation`.
        time: The time at the current snapshot. Changes when :any:`Snapshot.update` is called.
        config: The configuration array at the current snapshot. Changes when
            :any:`Snapshot.update` is called.
    """

    def __init__(self) -> None:
        """Init constructor for the base class.

        Parameters can be passed in here, and any attributes that can be defined
        without the parent :any:`Simulation` object can be instantiated here, such as
        :any:`update_time`.
        """
        self.simulation = None
        self.update_time = 0.1
        self.time = None
        self.config = None

    def initialize(self) -> None:
        """Method which is called once during :any:`add_snapshot`.

        Any initialization that requires accessing the data in :any:`simulation`
        should go here.
        """
        if self.simulation is None:
            raise ValueError('self.simulation is None, cannot call self.initialize until using sim.add_snapshot')

    def update(self, index: Optional[int] = None) -> None:
        """Method which is called while :any:`Snapshot.simulation` is running.

        Args:
            index: An optional integer index. If present, the snapshot will use the
                data from configuration :any:`configs` ``[index]`` and time
                :any:`times` ``[index]``. Otherwise, the snapshot will use the current
                configuration :any:`config_array` and current time.
        """
        if self.simulation is None:
            raise ValueError('self.simulation is None, cannot call self.update until using sim.add_snapshot')
        if index is not None:
            self.time = self.simulation.times[index]
            self.config = self.simulation.configs[index]
        else:
            self.time = self.simulation.time
            self.config = self.simulation.config_array


class TimeUpdate(Snapshot):
    """Simple :any:`Snapshot` that prints the current time in the :any:`Simulation`.

    When calling :any:`Simulation.run`, if :any:`snapshots` is empty, then
    this object will get added to provide a basic progress update.
    """

    def update(self, index: Optional[int] = None) -> None:
        super().update(index)
        print(f'\r Time: {self.time:.3f}', end='\r')


class Plotter(Snapshot):
    """Base class for a :any:`Snapshot` which will make a plot.

    Gives the option to map states to categories, for an easy way to visualize
    relevant subsets of the states rather than the whole state set.
    These require an interactive matplotlib backend to work.

    Attributes:
        fig: The matplotlib figure that is created.
        ax: The matplotlib axis object that is created. Modifying properties
            of this object is the most direct way to modify the plot.
        yscale: The scale used for the yaxis, passed into ax.set_yscale.
        state_map: A function mapping states to categories, which acts as a filter
            to view a subset of the states or just one field of the states.
        categories: A list which holds the set ``{state_map(state)}`` for all states
            in :any:`state_list`.
        sort_by:
        _matrix: A (# states)x(# categories) matrix such that for the configuration
            array (indexed by states), ``matrix * config`` gives an array
            of counts of categories. Used internally to get counts of categories.
    """

    def __init__(self, state_map: Optional[Callable[[State], Any]]=None, update_time=0.5, yscale='linear',
                 sort_by: str = 'categories') -> None:
        """Initializes the :any:`Plotter`.

        Args:
            state_map: An optional function mapping states to categories.
            yscale: The scale used for the yaxis, passed into ax.set_yscale.
                Defaults to 'linear'.
        """
        self._matrix = None
        self.state_map = state_map
        self.update_time = update_time
        self.yscale = yscale
        self.sort_by = sort_by

    def _add_state_map(self, state_map):
        """An internal function called to update :any:`categories` and `_matrix`."""
        self.categories = []

        for state in self.simulation.state_list:
            if state_map(state) is not None and state_map(state) not in self.categories:
                self.categories.append(state_map(state))
        self.categories = natsorted(self.categories, key=lambda x: repr(x))

        categories_dict = {j: i for i, j in enumerate(self.categories)}
        self._matrix = np.zeros((len(self.simulation.state_list), len(self.categories)), dtype=np.int64)
        for i, state in enumerate(self.simulation.state_list):
            m = state_map(state)
            if m is not None:
                self._matrix[i, categories_dict[m]] += 1

    def initialize(self) -> None:
        """Initializes the plotter by creating a fig and ax."""
        # Only do matplotlib import when necessary
        super().initialize()
        from matplotlib import pyplot as plt
        self.fig, self.ax = plt.subplots()
        if self.state_map is not None:
            self._add_state_map(self.state_map)
        else:
            self.categories = self.simulation.state_list


class StatePlotter(Plotter):
    """:any:`Plotter` which produces a barplot of counts."""

    def initialize(self) -> None:
        """Initializes the barplot.

        If :any:`state_map` gets changed, call :any:`initialize` to update the barplot to
            show the new set :any:`categories`.
        """
        super().initialize()
        import seaborn as sns
        self.ax = sns.barplot(x=[str(c) for c in self.categories], y=np.zeros(len(self.categories)))
        # rotate the x-axis labels if any of the label strings have more than 2 characters
        if max([len(str(c)) for c in self.categories]) > 2:
            for tick in self.ax.get_xticklabels():
                tick.set_rotation(90)
        self.ax.set_yscale(self.yscale)
        if self.yscale in ['symlog', 'log']:
            self.ax.set_ylim(0, 2 * self.simulation.simulator.n)
        else:
            self.ax.set_ylim(0, self.simulation.simulator.n)

    def update(self, index: Optional[int] = None) -> None:
        """Update the heights of all bars in the plot."""
        super().update(index)
        if self._matrix is not None:
            heights = np.matmul(self.config, self._matrix)
        else:
            heights = self.config
        for i, rect in enumerate(self.ax.patches):
            rect.set_height(heights[i])

        self.ax.set_title(f'Time {self.time: .3f}')
        self.fig.tight_layout()
        self.fig.canvas.draw()


class HistoryPlotter(Plotter):
    """Plotter which produces a lineplot of counts over time."""

    def update(self, index: Optional[int] = None) -> None:
        """Make a new history plot."""
        super().update(index)
        self.ax.clear()
        if self._matrix is not None:
            df = pd.DataFrame(data=np.matmul(self.simulation.history.to_numpy(), self._matrix),
                              columns=self.categories,
                              index=self.simulation.history.index)
        else:
            df = self.simulation.history
        df.plot(ax=self.ax)

        self.ax.set_yscale(self.yscale)
        if self.yscale in ['symlog', 'log']:
            self.ax.set_ylim(0, 2 * self.simulation.simulator.n)
        else:
            self.ax.set_ylim(0, 1.1 * self.simulation.simulator.n)

        # rotate the x labels if they are time units
        if self.simulation.time_units:
            for tick in self.ax.get_xticklabels():
                tick.set_rotation(45)
        self.fig.tight_layout()
        self.fig.canvas.draw()



"""
Module for expression population protocols using CRN notation. Ideas and much code taken from
https://github.com/enricozb/python-crn.

The general syntax is

.. code-block:: python

    a, b, u = species('A B U')
    approx_majority = [
        a + b >> 2 * u,
        a + u >> 2 * a,
        b + u >> 2 * b,
    ]
    n = 10 ** 5
    init_config = {a: 0.51 * n, b: 0.49 * n}
    sim = Simulation(init_config=init_config, rule=approx_majority)

In other words, a list of reactions is treated by the ppsim library just like the other ways of specifying
population protocol transitions (the `rule` parameter in the constructor for :any:`Simulation`, which also
accepts a dict or a Python function).

More examples given in https://github.com/UC-Davis-molecular-computing/ppsim/tree/main/examples
"""

from __future__ import annotations  # needed for forward references in type hints

from collections import defaultdict
import copy
from enum import Enum
from typing import Union, Dict, Tuple, Set, Iterable, DefaultDict, List, Any
from dataclasses import dataclass
from xml.dom import minidom


def species(sp: Union[str, Iterable[str]]) -> Tuple[Specie, ...]:
    """
    Create a list of :any:`Specie` (Single species :any:`Expression`'s),
    or a single one.

    args:
        sp:
            An string or Iterable of strings representing the names of the species being created.
            If a single string, species names are interpreted as space-separated.

    Examples:

    .. code-block:: python

        w, x, y, z = species('W X Y Z')
        rxn = x + y >> z + w


    .. code-block:: python

        w, x, y, z = species(['W', 'X', 'Y', 'Z'])
        rxn = x + y >> z + w

    """
    species_list: List[str]
    if isinstance(sp, str):
        species_list = sp.split()
    else:
        species_list = [specie.strip() for specie in sp]

    # if len(species_list) == 1:
    #     return Specie(species_list[0])
    if len(species_list) != len(set(species_list)):
        raise ValueError(f'species_list {species_list} cannot contain duplicates.')

    return tuple(map(Specie, species_list))


SpeciePair = Tuple['Specie', 'Specie']  # forward annotations don't seem to work here
Output = Union[SpeciePair, Dict[SpeciePair, float]]


def replace_reversible_rxns(rxns: Iterable[Reaction]) -> List[Reaction]:
    """
    Args:
        rxns: list of :any:`Reaction`'s

    Returns: list of :any:`Reaction`'s, where every reversible reaction in `rxns` has been replaced by
        two irreversible reactions, and all others have been left as they are
    """
    new_rxns: List[Reaction] = []
    for rxn in rxns:
        if not rxn.reversible:
            new_rxns.append(rxn)
        else:
            forward_rxn = Reaction(reactants=rxn.reactants, products=rxn.products, k=rxn.rate_constant,
                                   rate_constant_units=rxn.rate_constant_units, reversible=False)
            reverse_rxn = Reaction(reactants=rxn.products, products=rxn.reactants,
                                   k=rxn.rate_constant_reverse,
                                   rate_constant_units=rxn.rate_constant_reverse_units, reversible=False)
            new_rxns.extend([forward_rxn, reverse_rxn])
    return new_rxns


# XXX: This algorithm currently uses the reactant *ordered* pair.
# We should think about the logic of that and see if it makes sense to collapse
# two reversed ordered pairs to a single unordered pair at this step,
# or whether that should be done explicitly by the user specifying transition_order='symmetric'.
def reactions_to_dict(reactions: Iterable[Reaction], n: int, volume: float) \
        -> Tuple[Dict[SpeciePair, Output], float]:
    """
    Returns dict representation of `reactions`, transforming unimolecular reactions to bimolecular,
    and converting rates to probabilities, also returning the max rate so the :any:`Simulation` knows
    how to scale time.

    Args:
        reactions: list of :any:`Reaction`'s
        n: the population size, necessary for rate conversion
        volume: parameter as defined in Gillespie algorithm

    Returns:
        (transitions_dict, max_rate), where `transitions_dict` is the dict representation of the transitions,
        and `max_rate` is the maximum rate for any pair of reactants,
        i.e., if we have reactions (a + b >> c + d).k(2) and (a + b >> x + y).k(3),
        then the ordered pair (a,b) has rate 2+3 = 5
    """
    reactions = list(copy.deepcopy(reactions))

    # add reverse reactions explicitly
    reversible_reactions = [reaction for reaction in reactions if reaction.reversible]
    for reaction in reversible_reactions:
        reversed_reaction = Reaction(reactants=reaction.products,
                                     products=reaction.reactants,
                                     k=reaction.rate_constant_reverse,
                                     rate_constant_units=reaction.rate_constant_reverse_units,
                                     reversible=False)
        reactions.append(reversed_reaction)
        reaction.reversible = False

    # Make a copy of reactions because this conversion will mutate the reactions
    reactions = convert_unimolecular_to_bimolecular_and_flip_reactant_order(reactions, n, volume)

    # for each ordered pair of reactants, calculate sum of rate constants across all reactions with those reactants
    # get in stochastic units no matter how they're encoded in the reactions
    reactant_pair_rates: DefaultDict[SpeciePair, float] = defaultdict(int)
    for reaction in reactions:
        if not reaction.is_bimolecular():
            raise ValueError(f'all reactions must have exactly two reactants, violated by {reaction}')
        if not reaction.num_products() == 2:
            raise ValueError(f'all reactions must have exactly two products, violated by {reaction}')
        reactants = reaction.reactants_if_bimolecular()
        reactant_pair_rates[reactants] += reaction.rate_constant_stochastic

    # divide all rate constants by the max rate (per reactant pair) to help turn them into probabilities
    max_rate = max(reactant_pair_rates.values())

    # add one randomized transition per reaction
    transitions: Dict[SpeciePair, Output] = {}
    for reaction in reactions:
        prob = reaction.rate_constant_stochastic / max_rate
        reactants = reaction.reactants_if_bimolecular()
        products = reaction.products_if_exactly_two()
        if reactants not in transitions:
            # should we be worried about floating-point error here?
            # I hope not since dividing a float by itself should always result in 1.0.
            transitions[reactants] = products if prob == 1.0 else {products: prob}
        else:
            # if we calculated probabilities correctly above, if we assigned a dict entry to be non-randomized
            # (since prob == 1.0 above), we should not encounter another reaction with same reactants
            output = transitions[reactants]
            assert isinstance(output, dict)
            output[products] = prob

    # assert that each possible input for transitions has output probabilities summing to 1
    for reactants, outputs in transitions.items():
        if isinstance(outputs, dict):
            sum_probs = sum(prob for prob in outputs.values())
            assert sum_probs <= 1.0, f'sum_probs exceeds 1: {sum_probs}'

    return transitions, max_rate


def convert_unimolecular_to_bimolecular_and_flip_reactant_order(reactions: Iterable[Reaction], n: int,
                                                                volume: float) \
        -> List[Reaction]:
    """Process all reactions before being added to the dictionary.

    bimolecular reactions have their rates multiplied by the corrective factor (n-1) / (2 * volume).
    Bimolecular reactions with two different reactants are added twice, with their reactants in both orders.
    """

    # gather set of all species together
    all_species: Set[Specie] = set()
    for reaction in reactions:
        all_species.update(reaction.get_species())

    converted_reactions = []
    for reaction in reactions:
        if reaction.num_reactants() != reaction.num_products():
            raise ValueError(
                f'each reaction must have same number of reactants and products, violated by {reaction}')
        if reaction.is_bimolecular():
            # Corrective factor to reaction rate
            reaction.rate_constant *= (n - 1) / (2 * volume)
            converted_reactions.append(reaction)
            # Add a flipped copy if the reaction has two different reactants
            reactants = reaction.reactants_if_bimolecular()
            if len(set(reactants)) > 1:
                flipped_reaction = copy.copy(reaction)
                flipped_reaction.reactants = Expression([reactants[1], reactants[0]])
                assert flipped_reaction.reactants_if_bimolecular != reactants
                converted_reactions.append(flipped_reaction)

        elif reaction.is_unimolecular():
            # for each unimolecular reaction R -->(k) P and each species S,
            # add a bimolecular reaction R+S -->(k) P+S
            reactant = reaction.reactant_if_unimolecular()
            product = reaction.product_if_unique()
            bimolecular_implementing_reactions = [(reactant + s >> product + s).k(reaction.rate_constant)
                                                  for s in all_species]

            converted_reactions.extend(bimolecular_implementing_reactions)
        else:
            raise ValueError(f'each reaction must have exactly one or two reactants, violated by {reaction}')
    return converted_reactions


@dataclass(frozen=True)
class Specie:
    name: str

    def __add__(self, other: Union[Specie, Expression]) -> Expression:
        if isinstance(other, Expression):
            return other + Expression([self])
        elif isinstance(other, Specie):
            return Expression([self]) + Expression([other])

        raise NotImplementedError()

    __radd__ = __add__

    def __rshift__(self, other: Union[Specie, Expression]) -> Reaction:
        return Reaction(self, other)

    def __rrshift__(self, other: Union[Specie, Expression]) -> Reaction:
        return Reaction(other, self)

    def __or__(self, other: Union[Specie, Expression]) -> Reaction:
        return Reaction(self, other, reversible=True)

    def __mul__(self, other: int) -> Expression:
        if isinstance(other, int):
            return other * Expression([self])
        else:
            raise NotImplementedError()

    def __rmul__(self, other: int) -> Expression:
        if isinstance(other, int):
            return other * Expression([self])
        else:
            raise NotImplementedError()

    def __str__(self) -> str:
        return self.name

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, Specie):
            return NotImplemented
        return self.name < other.name

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Specie):
            return NotImplemented
        return self.name == other.name

    __req__ = __eq__


@dataclass(frozen=True)
class Expression:
    """
    Class used for very basic symbolic manipulation of left/right hand
    side of stoichiometric equations. Not very user friendly; users should
    just use the `species` functions and manipulate those to get their
    reactions.
    """

    species: List[Specie]
    """ordered list of species in expression, e.g, A+A+B is [A,A,B]"""

    def __add__(self, other: Expression) -> Expression:
        if isinstance(other, Expression):
            species_copy = list(self.species)
            species_copy.extend(other.species)
            return Expression(species_copy)
        else:
            raise NotImplementedError()

    def __rmul__(self, coeff: int) -> Expression:
        if isinstance(coeff, int):
            species_copy = []
            for _ in range(coeff):
                species_copy.extend(self.species)
            return Expression(species_copy)
        else:
            raise NotImplementedError()

    __mul__ = __rmul__

    def __rshift__(self, expr: Union[Specie, Expression]) -> Reaction:
        return Reaction(self, expr)

    def __or__(self, other: Union[Specie, Expression]) -> Reaction:
        return Reaction(self, other, reversible=True)

    def __str__(self) -> str:
        return ' + '.join(s.name for s in self.species)

    def __len__(self) -> int:
        return len(self.species)

    def get_species(self) -> Set[Specie]:
        """
        Returns the set of species in this expression, not their
        coefficients.
        """
        return set(self.species)


avogadro = 6.02214076e23


def concentration_to_count(concentration: float, volume: float) -> int:
    """

    Args:
        concentration: units of M (molar) = moles / liter
        volume: units of liter

    Returns:
        count of molecule with `concentration` in `volume`
    """
    return round(avogadro * concentration * volume)


class RateConstantUnits(Enum):
    stochastic = 'stochastic'
    """Units of L/s. Multiple by Avogadro's number to convert to mass-action units."""

    mass_action = 'mass_action'
    """Units of /M/s. Divide by Avogadro's number to convert to stochastic units."""


@dataclass
class Reaction:
    """
    Representation of a stoichiometric reaction using a pair of Expressions,
    one for the reactants and one for the products.
    """

    reactants: Expression
    """The left side of species in the reaction."""

    products: Expression
    """The right side of species in the reaction."""

    rate_constant: float = 1
    """Rate constant of forward reaction."""

    rate_constant_reverse: float = 1
    """Rate constant of reverse reaction (only used if :py:data:`Reaction.reversible` is true)."""

    rate_constant_units: RateConstantUnits = RateConstantUnits.stochastic
    """Units of forward rate constant."""

    rate_constant_reverse_units: RateConstantUnits = RateConstantUnits.stochastic
    """Units of reverse rate constant."""

    reversible: bool = False
    """Whether reaction is reversible, i.e. products &rarr; reactants is a reaction also."""

    def __init__(self, reactants: Union[Specie, Expression], products: Union[Specie, Expression],
                 k: float = 1, r: float = 1,
                 rate_constant_units: RateConstantUnits = RateConstantUnits.stochastic,
                 rate_constant_reverse_units: RateConstantUnits = RateConstantUnits.stochastic,
                 reversible: bool = False) -> None:
        """
        Args:
            reactants: left side of species in the reaction
            products: right side of species in the reaction
            k: Rate constant of forward reaction
            r: Rate constant of reverse reaction (only used if :py:data:`Reaction.reversible` is true
            rate_constant_units: Units of forward rate constant
            rate_constant_reverse_units: Units of reverse rate constant
            reversible: Whether reaction is reversible
        """
        if not (isinstance(reactants, Specie) or isinstance(reactants, Expression)):
            raise ValueError(
                "Attempted construction of reaction with type of reactants "
                f"as {type(reactants)}. Type of reactants must be Species "
                "or Expression")
        if not (isinstance(products, Specie) or isinstance(products, Expression)):
            raise ValueError(
                "Attempted construction of products with type of products "
                f"as {type(products)}. Type of products must be Species "
                "or Expression")

        if isinstance(reactants, Specie):
            reactants = Expression([reactants])
        if isinstance(products, Specie):
            products = Expression([products])

        if len(reactants) == 0:
            raise ValueError('reactants cannot be empty')

        if len(products) == 0:
            raise ValueError('products cannot be empty')

        self.reactants = reactants
        self.products = products
        self.rate_constant = float(k)
        self.rate_constant_reverse = float(r)
        self.rate_constant_units = rate_constant_units
        self.rate_constant_reverse_units = rate_constant_reverse_units
        self.reversible = reversible

    def is_unimolecular(self) -> bool:
        """
        Returns: true if there is one reactant
        """
        return self.num_reactants() == 1

    def is_bimolecular(self) -> bool:
        """
        Returns: true if there are two reactants
        """
        return self.num_reactants() == 2

    def symmetric(self) -> bool:
        """
        Returns: true if there are two reactants that are the same species
        """
        return self.num_reactants() == 2 and self.reactants.species[0] == self.reactants.species[1]

    def symmetric_products(self) -> bool:
        """
        Returns: true if there are two products that are the same species
        """
        return self.num_products() == 2 and self.products.species[0] == self.products.species[1]

    def num_reactants(self) -> int:
        """
        Returns: number of reactants
        """
        return len(self.reactants)

    def num_products(self) -> int:
        """
        Returns: number of products
        """
        return len(self.products)

    def is_conservative(self) -> bool:
        """
        Returns: true if number of reactants equals number of products
        """
        return self.num_reactants() == self.num_products()

    def reactant_if_unimolecular(self) -> Specie:
        """
        Returns: unique reactant if there is only one
        Raises: ValueError if there are multiple reactants
        """
        if self.is_unimolecular():
            return self.reactants.species[0]
        else:
            raise ValueError(f'reaction {self} is not unimolecular')

    def product_if_unique(self) -> Specie:
        """
        Returns: unique product if there is only one
        Raises: ValueError if there are multiple products
        """
        if self.num_products() == 1:
            return self.products.species[0]
        else:
            raise ValueError(f'reaction {self} does not have exactly one product')

    def reactants_if_bimolecular(self) -> Tuple[Specie, Specie]:
        """
        Returns: pair of reactants if there are exactly two
        Raises: ValueError if there are not exactly two reactants
        """
        if self.is_bimolecular():
            return self.reactants.species[0], self.reactants.species[1]
        else:
            raise ValueError(f'reaction {self} is not bimolecular')

    def reactant_names_if_bimolecular(self) -> Tuple[str, str]:
        """
        Returns: pair of reactant names if there are exactly two
        Raises: ValueError if there are not exactly two reactants
        """
        r1, r2 = self.reactants_if_bimolecular()
        return r1.name, r2.name

    def products_if_exactly_two(self) -> Tuple[Specie, Specie]:
        """
        Returns: pair of products if there are exactly two
        Raises: ValueError if there are not exactly two products
        """
        if self.num_products() == 2:
            return self.products.species[0], self.products.species[1]
        else:
            raise ValueError(f'reaction {self} does not have exactly two products')

    def product_names_if_exactly_two(self) -> Tuple[str, str]:
        """
        Returns: pair of product names if there are exactly two
        Raises: ValueError if there are not exactly two products
        """
        p1, p2 = self.products_if_exactly_two()
        return p1.name, p2.name

    def __str__(self) -> str:
        rev_rate_str = f'({self.rate_constant_reverse})<' if self.reversible else ''
        return f"{self.reactants} {rev_rate_str}-->({self.rate_constant}) {self.products}"

    def __repr__(self) -> str:
        return (f"Reaction({repr(self.reactants)}, {repr(self.products)}, "
                f"{self.rate_constant})")

    @property
    def rate_constant_stochastic(self) -> float:
        """
        Returns: forward rate constant in stochastic units (converts from mass-action if necessary)
        """
        return self.rate_constant \
            if self.rate_constant_units == RateConstantUnits.stochastic \
            else self.rate_constant / avogadro

    @property
    def rate_constant_reverse_stochastic(self) -> float:
        """
        Returns: reverse rate constant in stochastic units (converts from mass-action if necessary)
        """
        return self.rate_constant_reverse \
            if self.rate_constant_reverse_units == RateConstantUnits.stochastic \
            else self.rate_constant_reverse / avogadro

    def k(self, coeff: float, units: RateConstantUnits = RateConstantUnits.stochastic) -> Reaction:
        """
        Changes the reaction coefficient to `coeff` and returns `self`.

        This is useful for including the rate constant during the construction
        of a reaction. For example

        .. code-block:: python

            x, y, z = species("X Y Z")
            rxns = [
                (x + y >> z).k(2.5),
                (z >> x).k(1.5),
                (z >> y).k(0.5)),
            ]

        args:
            coeff: float
                The new reaction coefficient
            units: float
                units of rate constant (default stochastic)
        """
        if self.is_unimolecular() and units == RateConstantUnits.mass_action:
            raise ValueError('cannot use mass-action rate constants on a unimolecular reaction')
        self.rate_constant_units = units
        self.rate_constant = coeff
        return self

    def r(self, coeff: float, units: RateConstantUnits = RateConstantUnits.stochastic) -> Reaction:
        """
        Changes the reverse reactionn reaction rate constant to `coeff` and returns `self`.

        This is useful for including the rate constant during the construction
        of a reaction. For example

        .. code-block:: python

            x, y, z = species("X Y Z")
            rxns = [
                (x + y >> z).k(2.5),
                (z >> x).k(1.5),
                (z >> y).k(0.5)),
            ]

        args:
            coeff: float
                The new reverse reaction rate constant
            units: float
                units of rate constant (default stochastic)
        """
        if self.num_products() == 1 and units == RateConstantUnits.mass_action:
            raise ValueError('cannot use mass-action rate constants on a unimolecular reaction; '
                             'this reaction has only one product, so its reverse is unimolecular')
        if not self.reversible:
            raise ValueError('cannot set r on an irreversible reaction')
        self.rate_constant_reverse_units = units
        self.rate_constant_reverse = coeff
        return self

    def get_species(self) -> Set[Specie]:
        """
        Return: the set of species present in the products and reactants.
        """
        return {
            *self.reactants.get_species(),
            *self.products.get_species()
        }


# example of StochKit format:
'''
<Model>
   <Description>Epidemic</Description>
   <NumberOfReactions>1</NumberOfReactions>
   <NumberOfSpecies>2</NumberOfSpecies>
   <ParametersList>
     <Parameter>
       <Id>c1</Id>
       <Expression>1.0</Expression>
     </Parameter>
   </ParametersList>
   <ReactionsList>
     <Reaction>
       <Id>R2</Id>
       <Description> A+B -> 2B </Description>
       <Type>mass-action</Type>
       <Rate>c1</Rate>
       <Reactants>
           <SpeciesReference id="A" stoichiometry="1"/>
           <SpeciesReference id="B" stoichiometry="1"/>
       </Reactants>
       <Products>
           <SpeciesReference id="B" stoichiometry="2"/>
       </Products>
     </Reaction>
  </ReactionsList>
  <SpeciesList>
     <Species>
       <Id>A</Id>
       <Description>Species #1</Description>
       <InitialPopulation>10000</InitialPopulation>
     </Species>
     <Species>
       <Id>B</Id>
       <Description>Species #2</Description>
       <InitialPopulation>1</InitialPopulation>
     </Species>
  </SpeciesList>
</Model>
'''


def species_in_rxns(rxns: Iterable[Reaction]) -> List[Specie]:
    """
    Args:
        rxns: iterable of :any:`Reaction`'s

    Returns: list of species (without repetitions) in :any:`Reaction`'s in `rxns`
    """
    species_set: Set[Specie] = set()
    species_list: List[Specie] = []
    for rxn in rxns:
        for sp in rxn.reactants.species + rxn.products.species:
            if sp not in species_set:
                species_set.add(sp)
                species_list.append(sp)
    return species_list


def gillespy2_format(init_config: Dict[Specie, int], rxns: Iterable[Reaction],
                     volume: float = 1.0) -> Any:
    """
    Create a gillespy2 Model object from a CRN description.

    Args:
        init_config: dict mapping each :any:`Specie` to its initial count
        rxns: reactions to translate to StochKit format
        volume: volume in liters
        name: name of the CRN

    Returns:
        An equivalent gillespy2 Model object
    """
    # requires package gillespy2 to be installed
    import gillespy2 # type: ignore

    rxns = replace_reversible_rxns(rxns)
    species_list = species_in_rxns(rxns)
    model = gillespy2.Model()

    init_config = defaultdict(int, init_config)

    gillespy2_species = {s: gillespy2.Species(name=s.name, initial_value=init_config[s]) for s in
                         species_list}
    model.add_species(list(gillespy2_species.values()))
    model.volume = volume
    rates = [gillespy2.Parameter(name='r' + str(i), expression=r.rate_constant) for i, r in enumerate(rxns)]
    model.add_parameter(rates)
    for rxn, rate in zip(rxns, rates):
        reactants = {gillespy2_species[s]: rxn.reactants.species.count(s) for s in
                     rxn.reactants.get_species()}
        # Divide rate by 2 in same-species bimolecular reaction because gillespy2 propensity would be x(x-1)
        if list(reactants.values()) == [2]:
            rate.expression = float(rate.expression) / 2
        products = {gillespy2_species[s]: rxn.products.species.count(s) for s in rxn.products.get_species()}
        model.add_reaction(gillespy2.Reaction(reactants=reactants, products=products, rate=rate))
    return model


def stochkit_format(init_config: Dict[Specie, int], rxns: Iterable[Reaction],
                    volume: float = 1.0, name: str = 'CRN') -> str:
    """

    Args:
        rxns: reactions to translate to StochKit format
        init_config: dict mapping each :any:`Specie` to its initial count
        volume: volume in liters
        name: name of the CRN

    Returns:
        string describing CRN in StochKit XML format
    """
    rxns = replace_reversible_rxns(rxns)
    species_list = species_in_rxns(rxns)

    root = minidom.Document()

    model = root.createElement('Model')
    root.appendChild(model)

    desc_node = root.createElement('Description')
    model.appendChild(desc_node)
    desc_text = root.createTextNode(name)
    desc_node.appendChild(desc_text)

    num_rxns_node = root.createElement('NumberOfReactions')
    model.appendChild(num_rxns_node)
    num_rxns_text = root.createTextNode(str(len(rxns)))
    num_rxns_node.appendChild(num_rxns_text)

    num_species_node = root.createElement('NumberOfSpecies')
    model.appendChild(num_species_node)
    num_species_text = root.createTextNode(str(len(species_list)))
    num_species_node.appendChild(num_species_text)

    species_node = root.createElement('SpeciesList')
    model.appendChild(species_node)
    for specie in species_list:
        specie_node = root.createElement('Species')
        species_node.appendChild(specie_node)

        id_node = root.createElement('Id')
        specie_node.appendChild(id_node)
        id_text = root.createTextNode(specie.name)
        id_node.appendChild(id_text)

        specie_description_node = root.createElement('Description')
        specie_node.appendChild(specie_description_node)
        specie_description_text = root.createTextNode(specie.name)
        specie_description_node.appendChild(specie_description_text)

        initial_population_node = root.createElement('InitialPopulation')
        specie_node.appendChild(initial_population_node)
        count = init_config.get(specie, 0)
        initial_population_text = root.createTextNode(str(count))
        initial_population_node.appendChild(initial_population_text)

    rxns_node = root.createElement('ReactionsList')
    model.appendChild(rxns_node)
    for idx, rxn in enumerate(rxns):
        rxn_node = root.createElement('Reaction')
        rxns_node.appendChild(rxn_node)

        id_node = root.createElement('Id')
        rxn_node.appendChild(id_node)
        id_text = root.createTextNode(f'R{idx}')
        id_node.appendChild(id_text)

        description_node = root.createElement('Description')
        rxn_node.appendChild(description_node)
        description_text = root.createTextNode(str(rxn))
        description_node.appendChild(description_text)

        type_node = root.createElement('Type')
        rxn_node.appendChild(type_node)
        type_text = root.createTextNode('mass-action')
        type_node.appendChild(type_text)

        rate_node = root.createElement('Rate')
        rxn_node.appendChild(rate_node)
        rate = rxn.rate_constant_stochastic
        if rxn.is_bimolecular():
            rate /= volume
        rate_text = root.createTextNode(f'{rate}')
        rate_node.appendChild(rate_text)

        # reactants
        reactants_node = root.createElement('Reactants')
        rxn_node.appendChild(reactants_node)
        for reactant in rxn.reactants.get_species():
            reactant_node = root.createElement('SpeciesReference')
            reactants_node.appendChild(reactant_node)
            reactant_node.setAttribute('id', reactant.name)
            reactant_node.setAttribute('stoichiometry', str(rxn.reactants.species.count(reactant)))

        # products
        products_node = root.createElement('Products')
        rxn_node.appendChild(products_node)
        for product in rxn.products.get_species():
            product_node = root.createElement('SpeciesReference')
            products_node.appendChild(product_node)
            product_node.setAttribute('id', product.name)
            product_node.setAttribute('stoichiometry', str(rxn.products.species.count(product)))

    stochkit_xml: str = root.toprettyxml(indent='  ')

    return stochkit_xml


def write_stochkit_file(filename: str, rxns: Iterable[Reaction], init_config: Dict[Specie, int],
                        volume: float = 1.0, name: str = 'CRN') -> None:
    """
    Write stochkit file
    Args:
        filename: name of file to write
        rxns: reactions to translate to StochKit format
        init_config: dict mapping each :any:`Specie` to its initial count
        volume: volume in liters
        name: name of the CRN
    """
    xml = stochkit_format(rxns, init_config, volume, name)
    with open(filename, 'w') as f:
        f.write(xml)



from distutils.core import setup
from Cython.Build import cythonize
import numpy as np
from os.path import join
from setuptools.extension import Extension

inc_path = np.get_include()
lib_path = join(np.get_include(), '..', '..', 'random', 'lib')

with open("README.md", 'r') as f:
    long_description = f.read()

distributions = Extension("ppsim.simulator",
                          sources=[join('', 'ppsim/simulator.pyx')],
                          include_dirs=[inc_path],
                          library_dirs=[lib_path],
                          libraries=['npyrandom']
                          )

with open("requirements.txt") as fp:
    install_requires = fp.read().strip().split("\n")

# this is ugly, but appears to be standard practice:
# https://stackoverflow.com/questions/17583443/what-is-the-correct-way-to-share-package-version-with-setup-py-and-the-package/17626524#17626524
def extract_version(filename: str):
    with open(filename) as f:
        lines = f.readlines()
    version_comment = '# version line; WARNING: do not remove or change this line or comment'
    for line in lines:
        if version_comment in line:
            idx = line.index(version_comment)
            line_prefix = line[:idx]
            parts = line_prefix.split('=')
            parts = [part.strip() for part in parts]
            version_str = parts[-1]
            version_str = version_str.replace('"', '')
            version_str = version_str.replace("'", '')
            version_str = version_str.strip()
            return version_str
    raise AssertionError(f'could not find version in {filename}')

version = extract_version('ppsim/__version__.py')
print(f'ppsim version = {version}')

setup(
    name="ppsim",
    packages=['ppsim'],
    version=version,
    author="Eric Severson and David Doty",
    description="A package for simulating population protocols.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    python_requires='>=3.7',
    url="https://github.com/UC-Davis-molecular-computing/ppsim",
    ext_modules=cythonize(distributions, compiler_directives={'language_level': "3"}),
    install_requires=install_requires
)



"""
A module for simulating population protocols.

The main class :any:`Simulation` is created with a description of the protocol and the initial condition,
and is responsible for running the simulation.

The general syntax is

.. code-block:: python

    a, b, u = 'A', 'B', 'U'
    approx_majority = {
        (a,b): (u,u),
        (a,u): (a,a),
        (b,u): (b,b),
    }
    n = 10 ** 5
    init_config = {a: 0.51 * n, b: 0.49 * n}
    sim = Simulation(init_config=init_config, rule=approx_majority)
    sim.run()
    sim.history.plot()

More examples given in https://github.com/UC-Davis-molecular-computing/ppsim/tree/main/examples

:py:meth:`time_trials` is a convenience function used for gathering data about the
convergence time of a protocol.
"""

import dataclasses
from dataclasses import dataclass
from datetime import timedelta
import math
from time import perf_counter
from typing import Union, Hashable, Dict, Tuple, Callable, Optional, List, Iterable, Set, Any
from natsort import natsorted
import numpy as np
import pandas as pd
from tqdm.auto import tqdm

from . import simulator
from ppsim.crn import Reaction, reactions_to_dict
from ppsim.snapshot import Snapshot, TimeUpdate

# TODO: these names are not showing up in the mouseover information
State = Hashable
Output = Union[Tuple[State, State], Dict[Tuple[State, State], float]]
TransitionFunction = Callable[[State, State], Output]
Rule = Union[TransitionFunction, Dict[Tuple[State, State], Output], Iterable[Reaction]]
"""Type representing transition rule for protocol. Is one of three types: TODO"""

ConvergenceDetector = Callable[[Dict[State, int]], bool]


# TODO: give other option for when the number of reachable states is large or unbounded
def state_enumeration(init_dist: Dict[State, int], rule: Callable[[State, State], Output]) -> Set[State]:
    """Finds all reachable states by breadth-first search.

    Args:
        init_dist: dictionary mapping states to counts
            (states are any hashable type, commonly NamedTuple or String)
        rule: function mapping a pair of states to either a pair of states
            or to a dictionary mapping pairs of states to probabilities

    Returns:
        a set of all reachable states
    """
    checked_states = set()
    unchecked_states = set(init_dist.keys())
    while len(unchecked_states) > 0:
        unchecked_state = unchecked_states.pop()
        if unchecked_state not in checked_states:
            checked_states.add(unchecked_state)
        for checked_state in checked_states:
            for new_states in [rule(checked_state, unchecked_state),
                               rule(unchecked_state, checked_state)]:
                if new_states is not None:
                    if isinstance(new_states, dict):
                        # if the output is a distribution
                        new_states = sum(new_states.keys(), ())
                    for new_state in new_states:
                        if new_state not in checked_states and new_state not in unchecked_states:
                            unchecked_states.add(new_state)
    return checked_states


# TODO: add seed
@dataclass
class Simulation:
    """Class to simulate a population protocol."""

    state_list: List[State]
    """A sorted list of all reachable states."""
    state_dict: Dict[State, int]
    """Maps states to their integer index to be used
            in array representations."""
    simulator: simulator.Simulator
    """An internal :any:`Simulator` object, whose methods actually
            perform the steps of the simulation."""
    configs: List[np.ndarray]
    """A list of all configurations that have been
            recorded during the simulation, as integer arrays."""
    time: float
    """The current time."""
    times: List[Union[float, timedelta]]
    """A list of all the corresponding times for configs."""
    steps_per_time_unit: float
    """Number of simulated interactions per time unit."""
    time_units: Optional[str]
    """The units that time is in."""
    continuous_time: bool
    """Whether continuous time is used. The regular discrete
            time model considers :any:`steps_per_time_unit` steps 
            to be 1 unit of time.
            The continuous time model is a poisson process, with expected
            :any:`steps_per_time_unit` steps per 1 unit of time."""
    column_names: Union[pd.MultiIndex, List[str]]
    """Columns representing all states for pandas dataframe.
            If the State is a tuple, NamedTuple, or dataclass, this will be a
            pandas MultiIndex based on the various fields.
            Otherwise it is list of str(State) for each State."""
    snapshots: List[Snapshot]
    """A list of :any:`Snapshot` objects, which get
            periodically called during the running of the simulation to give live
            updates."""
    rng: np.random.Generator
    """A numpy random generator used to sample random variables outside the
            cython code."""
    seed: Optional[int]
    """The optional integer seed used for rng and inside cython code."""

    def __init__(self, init_config: Dict[State, int], rule: Rule, simulator_method: str = "MultiBatch",
                 transition_order: str = "symmetric", seed: Optional[int] = None,
                 volume: Optional[float] = None, continuous_time: bool = False, time_units: Optional[str] = None,
                 **kwargs):
        """Initialize a Simulation.

        Args:
            init_config: The starting configuration, as a
                dictionary mapping states to counts.
            rule (:any:`Rule`): A representation of the transition rule. The first two options are
                a dictionary, whose keys are tuples of 2 states and values are their
                outputs, or a function which takes pairs of states as input. For a
                deterministic transition function, the output is a tuple of 2 states.
                For a probabilistic transition function, the output is a dictionary
                mapping tuples of states to probabilities. Inputs that are not present
                in the dictionary, or return None from the function, are interpreted as
                null transitions that return the same pair of states as the output.
                The third option is a list of :any:`Reaction` objects describing a CRN,
                which will be parsed into an equivalent population protocol.
            simulator_method: Which Simulator method to use, either ``'MultiBatch'``
                or ``'Sequential'``.

                ``'MultiBatch'``:
                    :any:`SimulatorMultiBatch` does O(sqrt(n)) interaction steps in parallel
                    using batching, and is much faster for large population sizes and
                    relatively small state sets.
                ``'Sequential'``:
                    :any:`SimulatorSequentialArray` represents the population as an array of
                    agents, and simulates each interaction step by choosing a pair of agents
                    to update. Defaults to 'MultiBatch'.
            transition_order: Should the rule be interpreted as being symmetric,
                either ``'asymmetric'``, ``'symmetric'``, or ``'symmetric_enforced'``.
                Defaults to 'symmetric'.

                ``'asymmetric'``:
                    Ordering of the inputs matters, and all inputs not
                    explicitly given as assumed to be null interactions.

                ``'symmetric'``:
                    The input pairs are interpreted as unordered. If rule(a, b)
                    returns None, while rule(b, a) has a non-null output, then the
                    output of rule(a, b) is assumed to be the same as rule(b, a).
                    If rule(a, b) and rule(b, a) are each given, there is no effect.
                    Asymmetric interactions can be explicitly included this way.

                ``'symmetric_enforced'``:
                    The same as symmetric, except that if rule(a, b)
                    and rule(b, a) are non-null and do not give the same set of outputs,
                    a ValueError is raised.
            seed: An optional integer used as the seed for all pseudorandom number
                generation. Defaults to None.
            volume: If a list of :any:`Reaction` objects is given for a CRN, then
                the parameter volume can be passed in here. Defaults to None.
                If None, the volume will be assumed to be the population size n.
            continuous_time: Whether continuous time is used. Defaults to False.
                If a CRN as a list of reactions is passed in, this will be set to True.
            time_units: An optional string given the units that time is in. Defaults to None.
                This must be a valid string to pass as unit to pandas.to_timedelta.
            **kwargs: If `rule` is a function, any extra function parameters are passed in here,
                beyond the first two arguments representing the two agents. For example, if `rule` is
                defined:

                .. code-block:: python

                    def rule(sender: int, receiver: int, threshold: int) -> Tuple[int, int]:
                        if sender + receiver > threshold:
                            return 0, 0
                        else:
                            return sender, receiver+1

                To use a threshold of 20 in each interaction, in the :any:`Simulation` constructor, use

                .. code-block:: python

                    sim = Simulation(init_config, rule, threshold=20)

        """
        self.seed = seed
        self.rng = np.random.default_rng(seed)
        self.n = sum(init_config.values())
        self.steps_per_time_unit = self.n
        self.time_units = time_units
        self.continuous_time = continuous_time
        # if rule is iterable of Reactions from the crn module, then convert to dict
        rule_is_reaction_iterable = True
        try:
            for reaction in rule:
                if not isinstance(reaction, Reaction):
                    rule_is_reaction_iterable = False
                    break
        except:
            # might end up here if rule is not even iterable, e.g., is a function
            rule_is_reaction_iterable = False
        if rule_is_reaction_iterable:
            if volume is None:
                volume = self.n
            rule, rate_max = reactions_to_dict(rule, self.n, volume)
            self.steps_per_time_unit *= rate_max
            # Default to continuous time for lists of reactions
            self.continuous_time = True

        self._rule = rule
        self._rule_kwargs = kwargs

        # Get a list of all reachable states, use the natsort library to put in a nice order.
        if type(self._rule) == dict:
            # If the rule is a dict, we can loop over the entries to get all states
            states = []
            for input, output in self._rule.items():
                states.extend(input)
                if type(output) == dict:
                    for pair in output.keys():
                        states.extend(pair)
                else:
                    states.extend(output)
            state_list = list(set(states))
        else:
            # Otherwise, we use breadth-first search to find all reachable states
            state_list = list(state_enumeration(init_config, self.rule))
        # We use the natsorted library to put state_list in a reasonable order
        self.state_list = natsorted(state_list, key=lambda x: repr(x))
        self.state_dict = {state: i for i, state in enumerate(self.state_list)}

        if simulator_method.lower() == 'multibatch':
            self._method = simulator.SimulatorMultiBatch
        elif simulator_method.lower() == 'sequential':
            self._method = simulator.SimulatorSequentialArray
        else:
            raise ValueError('simulator_method must be multibatch or sequential')
        self._transition_order = transition_order
        self.initialize_simulator(self.array_from_dict(init_config))

        # Check an arbitrary state to see if it has fields.
        # This will be true for either a tuple, NamedTuple, or dataclass.
        state = next(iter(init_config.keys()))
        # Check for dataclass.
        if dataclasses.is_dataclass(state):
            field_names = [field.name for field in dataclasses.fields(state)]
            tuples = [dataclasses.astuple(state) for state in self.state_list]
        else:
            # Check for NamedTuple.
            field_names = getattr(state, '_fields', None)
            tuples = self.state_list
        # Check also for tuple.
        if (field_names and len(field_names) > 1)\
                or (isinstance(state, tuple) and len(state) > 1):
            # Make a MultiIndex only if there are multiple fields
            self.column_names = pd.MultiIndex.from_tuples(tuples, names=field_names)
        else:
            self.column_names = [str(i) for i in self.state_list]
        self.configs = []
        self.times = []
        self.time = 0
        self.add_config()
        # private history dataframe is initially empty, updated by the getter of property self.history
        self._history = pd.DataFrame(data=self.configs, index=pd.Index(self.times_in_units(self.times)),
                                     columns=self.column_names)
        self.snapshots = []

    def rule(self, a, b):
        """The rule, as a function of two input states."""
        # If the input rule was a dict
        if type(self._rule) == dict:
            if (a, b) in self._rule:
                return self._rule[(a, b)]
        # If the input rule was a function, with possible kwargs
        elif callable(self._rule):
            # Make a fresh copy in the case of a dataclass in case the function mutates a, b
            if dataclasses.is_dataclass(a):
                a, b = dataclasses.replace(a), dataclasses.replace(b)
            output = self._rule(a, b, **self._rule_kwargs)
            # If function just mutates a, b but doesn't return, then return new a, b values
            return (a, b) if output is None else output
        else:
            raise TypeError("rule must be either a dict or a callable.")

    def initialize_simulator(self, config):
        """Build the data structures necessary to instantiate the :any:`Simulator` class.

        Args:
            config: The config array to instantiate :any:`Simulator`.
        """
        q = len(self.state_list)
        delta = np.zeros((q, q, 2), dtype=np.intp)
        null_transitions = np.zeros((q, q), dtype=bool)
        random_transitions = np.zeros((q, q, 2), dtype=np.intp)
        random_outputs = []
        transition_probabilities = []
        for i, a in enumerate(self.state_list):
            for j, b in enumerate(self.state_list):
                output = self.rule(a, b)
                # when output is a distribution
                if type(output) == dict:
                    s = sum(output.values())
                    assert s <= 1, "The sum of output probabilities must be <= 1."
                    # ensure probabilities sum to 1
                    if 1 - s:
                        if (a, b) in output.keys():
                            output[(a, b)] += 1 - s
                        else:
                            output[(a, b)] = 1 - s
                    if len(output) == 1:
                        # distribution only had 1 output, not actually random
                        output = next(iter(output.keys()))
                    else:
                        # add (number of outputs, index to outputs)
                        random_transitions[i, j] = (len(output), len(random_outputs))
                        for (x, y) in output.keys():
                            random_outputs.append((self.state_dict[x], self.state_dict[y]))
                        transition_probabilities.extend(list(output.values()))
                if output is None or set(output) == {a, b}:
                    null_transitions[i, j] = True
                    delta[i, j] = (i, j)
                elif type(output) == tuple:
                    delta[i, j] = (self.state_dict[output[0]], self.state_dict[output[1]])
        random_outputs = np.array(random_outputs, dtype=np.intp)
        transition_probabilities = np.array(transition_probabilities, dtype=float)

        if self._transition_order.lower() in ['symmetric', 'symmetric_enforced']:
            for i in range(q):
                for j in range(q):
                    # Set the output for i, j to be equal to j, i if null
                    if null_transitions[i, j]:
                        null_transitions[i, j] = null_transitions[j, i]
                        delta[i, j] = delta[j, i]
                        random_transitions[i, j] = random_transitions[j, i]
                    # If i, j and j, i are both non-null, with symmetric_enforced, check outputs are equal
                    elif self._transition_order.lower() == 'symmetric_enforced' \
                            and not null_transitions[j, i]:
                        if sorted(delta[i, j]) != sorted(delta[j, i]) or \
                                random_transitions[i, j, 0] != random_transitions[j, i, 0]:
                            a, b = self.state_list[i], self.state_list[j]
                            raise ValueError(f'''Asymmetric interaction:
                                            {a, b} -> {self.rule(a, b)}
                                            {b, a} -> {self.rule(b, a)}''')

        self.simulator = self._method(config, delta, null_transitions,
                                      random_transitions, random_outputs, transition_probabilities, self.seed)

    def array_from_dict(self, d: Dict) -> np.ndarray:
        """Convert a configuration dictionary to an array.

        Args:
            d: A dictionary mapping states to counts.

        Returns:
            An array giving counts of all states, in the order of
            self.state_list.
        """

        a = np.zeros(len(self.state_list), dtype=np.int64)
        for k in d.keys():
            a[self.state_dict[k]] += d[k]
        return a

    def run(self, run_until: Union[float, ConvergenceDetector] = None,
            history_interval: Union[float, Callable[[float], float]] = 1.,
            stopping_interval: float = 1., timer: bool = True) -> None:
        """Runs the simulation.

        Can give a fixed amount of time to run the simulation, or a function that checks
        the configuration for convergence.

        Args:
            run_until: The stop condition. To run for a fixed amount of time, give
                a numerical value. To run until a convergence criterion, give a function
                mapping a configuration (as a dictionary mapping states to counts) to a
                boolean. The run will stop when the function returns True.
                Defaults to None. If None, the simulation will run until the configuration
                is silent (all transitions are null). This only works with the multibatch
                simulator method, if another simulator method is given, then using None will
                raise a ValueError.
            history_interval: The length to run the simulator before recording data,
                in current time units. Defaults to 1.
                This can either be a float, or a function that takes the current time and
                and returns a float.
            stopping_interval: The length to run the simulator before checking for the stop
                condition.
            timer: If True, and there are no other snapshot objects, a default :any:`TimeUpdate`
                snapshot will be created to print updates with the current time.
                Defaults to True.
        """
        if len(self.snapshots) == 0 and timer is True:
            self.add_snapshot(TimeUpdate())

        end_time = None
        # stop_condition() returns True when it is time to stop
        if run_until is None:
            if type(self.simulator) != simulator.SimulatorMultiBatch:
                raise ValueError('Running until silence only works with multibatch simulator.')

            def stop_condition():
                return self.simulator.silent
        elif type(run_until) is float or type(run_until) is int:
            end_time = self.time + run_until

            def stop_condition():
                return self.time >= end_time
        elif callable(run_until):

            def stop_condition():
                return run_until(self.config_dict)
        else:
            raise TypeError('run_until must be a float, int, function, or None.')

        # Stop if stop_condition is already met
        if stop_condition():
            return

        def get_next_history_time():
            # Get the next time that will be recorded to self.times and self.history
            if callable(history_interval):
                length = history_interval(self.time)
            else:
                length = history_interval
            if length <= 0:
                raise ValueError('history_interval must always be strictly positive.')
            return self.time + length

        if stopping_interval <= 0:
            raise ValueError('stopping_interval must always be strictly positive.')

        next_history_time = get_next_history_time()

        def get_next_time():
            # Get the next time simulator will run until
            t = min(next_history_time, self.time + stopping_interval)
            if end_time is not None:
                t = min(t, end_time)
            return t

        next_time = get_next_time()
        # The next step that the simulator will be run until, which corresponds to parallel time next_time
        next_step = self.time_to_steps(next_time)

        for snapshot in self.snapshots:
            snapshot.next_snapshot_time = perf_counter() + snapshot.update_time

        # add max_wall_clock to be the minimum snapshot update time, to put a time bound on calls to simulator.run
        max_wallclock_time = [min([s.update_time for s in self.snapshots])] if len(self.snapshots) > 0 else []
        while stop_condition() is False:
            if self.time >= next_time:
                next_step += self.time_to_steps(get_next_time() - next_time)
                next_time = get_next_time()
            current_step = self.simulator.t
            self.simulator.run(next_step, *max_wallclock_time)
            if self.simulator.t == next_step:
                self.time = next_time
            elif self.simulator.t < next_step:
                # simulator exited early from hitting max_wallclock_time
                # add a fraction of the time until next_time equal to the fractional progress made by simulator
                self.time += (next_time - self.time) * (self.simulator.t - current_step) / (next_step - current_step)
            else:
                raise RuntimeError('The simulator ran to step {self.simulator.t} past the next step {next_step}.')
            if self.time >= next_history_time:
                assert self.time == next_history_time, \
                    f'self.time = {self.time} overshot next_history_time = {next_history_time}'
                self.add_config()
                next_history_time = get_next_history_time()
            for snapshot in self.snapshots:
                if perf_counter() >= snapshot.next_snapshot_time:
                    snapshot.update()
                    snapshot.next_snapshot_time = perf_counter() + snapshot.update_time
        # add the final configuration if it wasn't already recorded
        if self.time > self.times[-1]:
            self.add_config()
        # final updates for all snapshots
        for snapshot in self.snapshots:
            snapshot.update()

        if len(self.snapshots) == 1 and type(self.snapshots[0]) is TimeUpdate:
            self.snapshots.pop()
            print()

    @property
    def reactions(self) -> str:
        """
        A string showing all non-null transitions in reaction notation.

        Each reaction is separated by newlines, so that ``print(self.reactions)`` will display all reactions.
        Only works with simulator method multibatch, otherwise will raise a ValueError.
        """
        if type(self.simulator) != simulator.SimulatorMultiBatch:
            raise ValueError('reactions must be defined by multibatch simulator.')
        w = max([len(str(state)) for state in self.state_list])
        reactions = [self._reaction_string(r, p, w) for (r, p) in
                     zip(self.simulator.reactions, self.simulator.reaction_probabilities)]
        return '\n'.join(reactions)

    @property
    def enabled_reactions(self) -> str:
        """
        A string showing all non-null transitions that are currently enabled.

        This can only check the current configuration in self.simulator.
        Each reaction is separated by newlines, so that ``print(self.enabled_reactions)``
        will display all enabled reactions.
        """
        if type(self.simulator) != simulator.SimulatorMultiBatch:
            raise ValueError('reactions must be defined by multibatch simulator.')
        w = max([len(str(state)) for state in self.state_list])
        self.simulator.get_enabled_reactions()

        reactions = []
        for i in range(self.simulator.num_enabled_reactions):
            r = self.simulator.reactions[self.simulator.enabled_reactions[i]]
            p = self.simulator.reaction_probabilities[self.simulator.enabled_reactions[i]]
            reactions.append(self._reaction_string(r, p, w))
        return '\n'.join(reactions)

    def _reaction_string(self, reaction, p: float = 1, w: int = 1) -> str:
        """A string representation of a reaction."""

        reactants = [self.state_list[i] for i in sorted(reaction[0:2])]
        products = [self.state_list[i] for i in sorted(reaction[2:])]
        s = '{0}, {1}  -->  {2}, {3}'.format(*[str(x).rjust(w) for x in reactants + products])
        if p < 1:
            s += f'      with probability {p}'
        return s

    # TODO: If this changes n, then the timescale must change
    def reset(self, init_config: Optional[Dict[State, int]] = None) -> None:
        """Reset the simulation.

        Args:
            init_config: The configuration to reset to. Defaults to None.
                If None, will use the old initial configuration.
        """
        if init_config is None:
            config = self.configs[0]
        else:
            config = np.zeros(len(self.state_list), dtype=np.int64)
            for k in init_config.keys():
                config[self.state_dict[k]] += init_config[k]
        self.configs = [config]
        self.times = [0]
        self.time = 0
        self._history = pd.DataFrame(data=self.configs, index=pd.Index(self.times, name='time'),
                                     columns=self._history.columns)
        self.simulator.reset(config)

    # TODO: If this changes n, then the timescale must change
    def set_config(self, config: Union[Dict[State, int], np.ndarray]) -> None:
        """Change the current configuration.

        Args:
            config: The configuration to change to. This can be a dictionary,
                mapping states to counts, or an array giving counts in the order
                of :any:`state_list`.
        """
        if type(config) is dict:
            config_array = self.array_from_dict(config)
        else:
            config_array = np.array(config, dtype=np.int64)
        self.simulator.reset(config_array, self.simulator.t)
        self.add_config()

    def time_to_steps(self, time: float) -> int:
        """Convert length of time into number of steps.

        Args:
            time: The amount of time to convert.
        """
        expected_steps = time * self.steps_per_time_unit
        if self.continuous_time:
            # In continuous time the number of steps is a poisson random variable
            # TODO: handle the case when expected_steps is larger than an int32 and numpy reports a ValueError
            return self.rng.poisson(expected_steps)
        else:
            # In discrete time we round up to the next step
            return math.ceil(expected_steps)

    @property
    def config_dict(self) -> Dict[State, int]:
        """The current configuration, as a dictionary mapping states to counts."""
        return {self.state_list[i]: self.simulator.config[i] for i in np.nonzero(self.simulator.config)[0]}

    @property
    def config_array(self) -> np.ndarray:
        """The current configuration in the simulator, as an array of counts.

        The array is given in the same order as self.state_list. The index of state s
        is self.state_dict[s].
        """
        return np.asarray(self.simulator.config)

    @property
    def history(self) -> pd.DataFrame:
        """A pandas dataframe containing the history of all recorded configurations."""
        h = len(self._history)
        if h < len(self.configs):
            new_history = pd.DataFrame(data=self.configs[h:], index=pd.Index(self.times_in_units(self.times[h:])),
                                       columns=self._history.columns)
            self._history = pd.concat([self._history, new_history])
            if self.time_units is None:
                if self.continuous_time:
                    self._history.index.name = 'time (continuous units)'
                else:
                    n = "n" if self.n == self.steps_per_time_unit else str(self.steps_per_time_unit)
                    self._history.index.name = f'time ({n} interactions)'
        return self._history

    @property
    def null_probability(self) -> float:
        """The probability the next interaction is null."""
        if type(self.simulator) != simulator.SimulatorMultiBatch:
            raise ValueError('null probability requires by multibatch simulator.')
        self.simulator.get_enabled_reactions()
        n = self.simulator.n
        return 1 - self.simulator.get_total_propensity() / (n * (n-1) / 2)

    def times_in_units(self, times: Iterable[float]) -> Iterable[Any]:
        """If :any:`time_units` is defined, convert time list to appropriate units."""
        if self.time_units:
            return pd.to_timedelta(times, unit=self.time_units)
        else:
            return times

    def add_config(self) -> None:
        """Appends the current simulator configuration and time."""
        self.configs.append(np.array(self.simulator.config))
        self.times.append(self.time)

    def set_snapshot_time(self, time: float) -> None:
        """Updates all snapshots to the nearest recorded configuration to a specified time.

        Args:
            time: The parallel time to update the snapshots to.
        """
        index = np.searchsorted(self.times, time)
        for snapshot in self.snapshots:
            snapshot.update(index=index)

    def set_snapshot_index(self, index: int) -> None:
        """Updates all snapshots to the configuration :any:`configs` ``[index]``.

        Args:
            index: The index of the configuration.
        """
        for snapshot in self.snapshots:
            snapshot.update(index=index)

    def add_snapshot(self, snap: "Snapshot") -> None:
        """Add a new :any:`Snapshot` to :any:`snapshots`.

        Args:
            snap: The :any:`Snapshot` object to be added.
        """
        snap.simulation = self
        snap.initialize()
        snap.update()
        self.snapshots.append(snap)

    def snapshot_slider(self, var: str = 'index') -> "widgets.interactive":
        """Returns a slider that updates all :any:`Snapshot` objects.

        Returns a slider from the ipywidgets library.

        Args:
            var: What variable the slider uses, either ``'index'`` or ``'time'``.
        """
        import ipywidgets as widgets
        if var.lower() == 'index':
            return widgets.interactive(self.set_snapshot_index,
                                       index=widgets.IntSlider(min=0,
                                                               max=len(self.times) - 1,
                                                               layout=widgets.Layout(width='100%'),
                                                               step=1))
        elif var.lower() == 'time':
            return widgets.interactive(self.set_snapshot_time,
                                       time=widgets.FloatSlider(min=self.times[0],
                                                                max=self.times[-1],
                                                                layout=widgets.Layout(width='100%'),
                                                                step=0.01))
        else:
            raise ValueError("var must be either 'index' or 'time'.")

    def sample_silence_time(self) -> float:
        """Starts a new trial from the initial distribution and return time until silence."""
        if type(self.simulator) != simulator.SimulatorMultiBatch:
            raise ValueError('silence time can only be found by multibatch simulator.')
        self.simulator.run_until_silent(np.array(self.configs[0]))
        return self.time

    def sample_future_configuration(self, time: float, num_samples: int = 100) -> pd.DataFrame:
        """Repeatedly samples the configuration at a fixed future time.

        Args:
            time: The amount of time ahead to sample the configuration.
            num_samples: The number of samples to get.

        Returns:
            A dataframe whose rows are the sampled configuration.
        """
        samples = []
        t = self.simulator.t
        for _ in tqdm(range(num_samples)):
            self.simulator.reset(np.array(self.configs[-1]), t)
            end_step = t + self.time_to_steps(time)
            self.simulator.run(end_step)
            samples.append(np.array(self.simulator.config))
        return pd.DataFrame(data=samples, index=pd.Index(range(num_samples), name='trial #'),
                            columns=self._history.columns)

    def __getstate__(self):
        """Returns information to be pickled."""
        # Clear _history such that it can be regenerated by self.history
        d = dict(self.__dict__)
        d['_history'] = pd.DataFrame(data=self.configs[0:1], index=pd.Index(self.times_in_units(self.times[0:1])),
                                     columns=self._history.columns)
        del d['simulator']
        return d

    def __setstate__(self, state) -> None:
        """Instantiates from the pickled state information."""
        self.__dict__ = state
        self.initialize_simulator(self.configs[-1])


def time_trials(rule: Rule, ns: List[int], initial_conditions: Union[Callable, List],
                convergence_condition: Optional[Callable] = None, convergence_check_interval: float = 0.1,
                num_trials: int = 100, max_wallclock_time: float = 60 * 60 * 24,

                **kwargs) -> pd.DataFrame:
    """Gathers data about the convergence time of a rule.

    Args:
        rule: The rule that is used to generate the :any:`Simulation` object.
        ns: A list of population sizes n to sample from.
            This should be in increasing order.
        initial_conditions: An initial condition is a dict mapping states to counts.
            This can either be a list of initial conditions, or a function mapping
            population size n to an initial condition of n agents.
        convergence_condition: A boolean function that takes a configuration dict
            as input and returns True if that configuration has converged.
            Defaults to None. If None, the simulation will run until silent
            (all transitions are null), and the data will be for silence time.
        convergence_check_interval: How often (in parallel time) the simulation will
            run between convergence checks. Defaults to 0.1.
            Smaller values give better resolution, but spend more time checking for
            convergence.
        num_trials: The maximum number of trials that will be done for each
            population size n, if there is sufficient time. Defaults to 100.
            If you want to ensure that you get the full num_trials samples for
            each value n, use a large value for time_bound.
        max_wallclock_time: A bound (in seconds) for how long this function will run.
            Each value n is given a time budget based on the time remaining, and
            will stop before doing num_trials runs when this time budget runs out.
            Defaults to 60 * 60 * 24 (one day).
        **kwargs: Other keyword arguments to pass into :any:`Simulation`.

    Returns:
        A pandas dataframe giving the data from each trial, with two columns
        ``'n''`` and ``'time'``. A good way to visualize this dataframe is using the
        seaborn library, calling ``sns.lineplot(x='n', y='time', data=df)``.
    """
    d = {'n': [], 'time': []}
    end_time = perf_counter() + max_wallclock_time
    if callable(initial_conditions):
        initial_conditions = [initial_conditions(n) for n in ns]
    for i in tqdm(range(len(ns))):
        sim = Simulation(initial_conditions[i], rule, **kwargs)
        t = perf_counter()
        time_limit = t + (end_time - t) / (len(ns) - i)
        j = 0
        while j < num_trials and perf_counter() < time_limit:
            j += 1
            sim.reset(initial_conditions[i])
            sim.run(convergence_condition, stopping_interval=convergence_check_interval, timer=False)
            d['n'].append(ns[i])
            d['time'].append(sim.time)

    return pd.DataFrame(data=d)
